; crt1.s  04-Apr-89  A.J.Travis

;
; Run-time support for tcc under Acorn/BBC MOS
;

;
; _trace() - trace/debug trap
; ---------------------------
; print contents of 6502 regs, and tcc pseudo-regs in hex
;
__trace
	sta asave	;save 6502 regs
	stx pr
	sty pr+1
	jsr osnewl	;start of reg display
	ldy #1		;6502 stack hi
	tsx		;6502 stack lo
	jsr ~hex2
	lda $101,x	;return addr lo
	ldy $102,x	;return addr hi
	tax
	jsr ~hex2
	ldx pr		;primary reg lo
	ldy pr+1	;primary reg hi
	jsr ~hex2
	ldx sr		;secondary reg lo
	ldy sr+1	;secondary reg hi
	jsr ~hex2
	ldx tr		;tertiary reg lo
	ldy tr+1	;tertiary reg hi
	jsr ~hex2
	ldx sp		;data stack pointer lo
	ldy sp+1	;data stack pointer hi
	jsr ~hex2
	ldy #0
	lda (sp),y	;top stack item lo
	tax
	iny
	lda (sp),y	;top stack item hi
	tay
	jsr ~hex2
	lda asave	;restore 6502 regs
	ldx pr
	ldy pr+1
	rts
;
; output hex number followed by a blank
;
~hex2
	tya
	jsr ~hex
~hex1
	txa
	jsr ~hex
	lda #' '	
	jsr oswrch
	rts
;
; output a as 2 hex digits
;
~hex
	pha
	lsr a
	lsr a
	lsr a
	lsr a
	jsr ~ahex
	pla
	jsr ~ahex
	rts
;
; output bits 0-3 of a as hex digit
;
~ahex
	and #$0f
	cmp #$0a
	clc
	bmi ~ahex1
	adc #$07
~ahex1
	adc #$30
	jmp oswrch	;use mos rts
;
; sign extend char in xy
; ----------------------
; sign bit in status reg on entry
;
ext
	bpl ~ext1
	ldy #$ff	;negative
	rts
~ext1
	ldy #0		;positive
	rts
;
; convert sp offset to address
; -------------------------------
; addr   if offset = 0
; addr_1 if offset = 2
; addr_2 if offset = 4
; ...
; addr_6 if offset = 12
; addr_b if offset < 256
; addr_w if offset >= 256
;
addr
	ldx sp
	ldy sp+1
	rts
addr_1
	ldx #2	
	jmp addr_b
addr_2
	ldx #4
	jmp addr_b
addr_3
	ldx #6
	jmp addr_b
addr_4
	ldx #8	
	jmp addr_b
addr_5
	ldx #10
	jmp addr_b
addr_6
	ldx #12
addr_b
	ldy #0
addr_w
	clc
	txa
	adc sp
	tax
	tya
	adc sp+1
	tay
	rts
;
; store indirect byte in x at addr on stack
; -----------------------------------------
; leaves signed char in xy as result
;
sind_b
	jsr pop
	ldy #0
	txa
	sta (sr),y
	jmp ext
;
; store indirect word in xy at addr on stack
;
sind_w
	jsr pop
	sty pr+1
	ldy #0
	txa
	sta (sr),y
	iny
	lda pr+1
	sta (sr),y
	tay
	rts
;
; load indirect byte from addr in xy
;
lind_b
	stx pr
	sty pr+1
	ldy #0
	lda (pr),y
	tax
	jmp ext
;
; load indirect word from addr in xy
;
lind_w
	stx pr
	sty pr+1
	ldy #0
	lda (pr),y	;word lo
	tax
	iny
	lda (pr),y	;word hi
	tay
	rts
;
; call subroutine on data stack
; -----------------------------
; return addr is on 6502 return stack
;
scall
	ldy #0
	lda (sp),y	;addr lo
	sta tr
	iny
	lda (sp),y	;addr hi
	sta tr+1
	jsr drop	;adjust data stack
	jmp (tr)	;return via rts
;
; xy = xy + 1
;
inc1
	inx
	bne ~inc11
	iny
~inc11
	rts
;
; xy = xy + 2
;
inc2	
	clc
	txa
	adc #2
	tax
	tya
	adc #0
	tay
	rts	
;
; xy = xy - 1
;
dec1
	dex
	cpx #$ff
	bne ~dec11
	dey
~dec11
	rts
;
; xy = xy - 2
;
dec2	
	sec
	txa
	sbc #2
	tax
	tya
	sbc #0
	tay
	rts	
;
; increment sp by byte in a
;
sinc_b
	clc
	adc sp
	sta sp
	lda #0
	adc sp+1
	sta sp+1
	rts
;
; increment sp by word in a + asave
;
sinc_w
	clc
	adc sp
	sta sp
	lda asave
	adc sp+1
	sta sp+1
	rts
;
; decrement sp by byte in a
;
sdec_b
	sta tr
	sec
	lda sp
	sbc tr
	sta sp
	lda sp+1
	sbc #0
	sta sp+1
	rts
;
; decrement sp by word in a + asave
;
sdec_w
	sta tr
	sec
	lda sp
	sbc tr
	sta sp
	lda sp+1
	sbc asave
	sta sp+1
	rts
;
; swap xy and sr
;
swap
	txa
	ldx sr
	sta sr
	tya
	ldy sr+1
	sta sr+1
	rts
;
; push xy to data stack
;
push
	sec		;sp = sp - 2
	lda sp
	sbc #2
	sta sp
	lda sp+1
	sbc #0
	sta sp+1
	sty pr+1	;save y
	tya
	ldy #1
	sta (sp),y	;old y in stack item hi
	dey
	txa
	sta (sp),y	;old x in stack item lo
	ldy pr+1	;restore y
	rts
;
; pop data stack to sr
;
pop
	sty pr+1	;save y
	ldy #0
	lda (sp),y
	sta sr
	iny
	lda (sp),y
	sta sr+1
	ldy pr+1	;restore y
	jmp drop
;
; drop top three stack items
;
drop3
	jsr drop
;
; drop top two stack items
;
drop2
	jsr drop
drop
	clc		;drop top stack item
	lda sp
	adc #2
	sta sp
	lda sp+1
	adc #0
	sta sp+1
	rts
;
; exchange xy and top stack item
;
xchange
	sty pr+1	;save y
	ldy #0
	lda (sp),y	;save stack item lo
	sta pr
	txa
	sta (sp),y	;old x in stack item lo
	iny
	lda (sp),y	;save stack item hi in x
	tax
	lda pr+1
	sta (sp),y	;old y in stack item hi
	txa
	tay		;old stack item hi in y
	ldx pr		;old stack item lo in x
	rts
;
; xy = xy * 2
;
scale2
	lda #2
;
; xy = xy * a
;
scale
	sta tr
	stx pr+2
	sty pr+3
	ldy #0
	sty pr
	sty pr+1
	ldy #16
~scale1
	asl pr
	rol pr+1
	rol pr+2
	rol pr+3
	bcc ~scale2
	clc
	lda tr
	adc pr
	sta pr
	lda #0
	adc pr+1
	sta pr+1
~scale2
	dey
	bne ~scale1
	ldx pr
	ldy pr+1
	rts
;
; xy = top stack item + xy
;
add
	jsr pop
	clc
	txa
	adc sr
	tax
	tya
	adc sr+1
	tay
	rts
;
; xy = top stack item - xy
;
sub
	jsr pop
	stx pr
	sty pr+1
	sec
	lda sr
	sbc pr
	tax
	lda sr+1
	sbc pr+1
	tay
	rts
;
; xy = top stack item * xy
;
mult
	jsr pop
	stx pr+2
	sty pr+3
	ldy #0
	sty pr
	sty pr+1
	ldy #16
~mult1
	asl pr
	rol pr+1
	rol pr+2
	rol pr+3
	bcc ~mult2
	clc
	lda sr
	adc pr
	sta pr
	lda sr+1
	adc pr+1
	sta pr+1
	lda #0
	adc pr+2
	sta pr+2
~mult2
	dey
	bne ~mult1
	ldx pr
	ldy pr+1
	rts
;
; xy = top stack item / xy (16-bit signed)
;
div
	jsr pop		;divisor in sr
	tya		;dividend sign
	eor sr+1
	pha		;sign of quotient
	tya		;test sign of xy
	bpl ~div1
	jsr neg
~div1
	bit sr+1	;test sign of sr
	bpl ~div2
	jsr ~negsr
~div2
	jsr ~udiv	;unsigned divide
	pla		;sign of quotient
	bpl ~div3
	jsr neg		;quotient in xy
~div3
	rts
;
; xy = top stack item / xy (16-bit unsigned)
;
udiv
	jsr pop
;
; xy = xy / sr (16-bit unsigned)
;
~udiv
	jsr ~ztrap	;trap division by zero
	stx tr
	sty tr+1
	lda sr
	sta pr
	lda sr+1
	sta pr+1
	ldy #0
	sty sr
	sty sr+1
	ldx #16
	asl pr
	rol pr+1
~udiv2
	rol sr
	rol sr+1
	sec
	lda sr
	sbc tr
	tay
	lda sr+1
	sbc tr+1
	bcc ~udiv3
	sty sr
	sta sr+1
~udiv3
	rol pr
	rol pr+1
	dex
	bne ~udiv2
	ldx pr
	ldy pr+1
	rts
;
; xy = top stack item % xy (16-bit signed)
;
mod
	jsr pop		;divisor in sr
	tya		;dividend sign
	pha		;save sign
	bpl ~mod1
	jsr neg
~mod1
	bit sr+1	;test sign of sr
	bpl ~mod2
	jsr ~negsr
~mod2
	jsr ~umod	;unsigned remainder
	pla		;sign of result
	bpl ~mod3
	jsr neg
~mod3
	rts
;
; xy = top stack item % xy (16-bit unsigned)
;
umod
	jsr pop
;
; xy = sr % xy (16-bit unsigned)
;
~umod
	jsr ~ztrap	;trap division by zero
	stx tr
	sty tr+1
	ldy #0		;clear pr
	sty pr
	sty pr+1
	ldx #16		;bit count
	asl sr
	rol sr+1
~umod1
	rol pr
	rol pr+1
	sec
	lda pr
	sbc tr
	tay
	lda pr+1
	sbc tr+1
	bcc ~umod2
	sty pr
	sta pr+1
~umod2
	rol sr
	rol sr+1
	dex
	bne ~umod1
	ldx pr
	ldy pr+1
	rts
;
; division by zero trap
;
~ztrap
	txa
	bne ~ztrap1
	tya
	bne ~ztrap1
	brk
	.byte 0
	.text "division by zero"
	.byte 0
~ztrap1
	rts
;
; negate sr
;
~negsr
	sec
	lda #0
	sbc sr
	sta sr
	lda #0
	sbc sr+1
	sta sr+1
	rts
;
; xy = top stack item | xy
;
or
	jsr pop
	txa
	ora sr
	tax
	tya
	ora sr+1
	tay
	rts
;
; xy = top stack item ^ xy
;
xor
	jsr pop
	txa
	eor sr
	tax
	tya
	eor sr+1
	tay
	rts
;
; xy = top stack item & xy
;
cand
	jsr pop
	txa
	and sr
	tax
	tya
	and sr+1
	tay
	rts
;
; right shift top stack item xy bits
;
casr
	jsr pop
	lda sr+1
	sta pr+1
	lda sr
	cpx #0
	beq ~asr2
~asr1
	lsr pr+1
	ror a
	dex
	bne ~asr1
~asr2
	tax
	ldy pr+1
	rts
;
; left shift top stack item xy bits
;
casl
	jsr pop
	lda sr
	sta pr
	lda sr+1
	cpx #0
	beq ~asl2
~asl1
	asl pr
	rol a
	dex
	bne ~asl1
~asl2
	tay
	ldx pr
	rts
;
; negate xy
;
neg
	clc
	txa
	eor #$ff
	adc #1
	tax
	tya
	eor #$ff
	adc #0
	tay
	rts
;
; relational operators return status in 6502 'z' flag
; ---------------------------------------------------
; xy != 0
;
nz
	cpx #0	
	bne ~nz1
	cpy #0
~nz1
	rts
;
; signed top stack item == xy
;
eq
	jsr ~cmp
	beq ~eq1
	ldx #0
	rts
~eq1
	ldx #1
	rts
;
; signed top stack item != xy
;
ne
	jsr ~cmp
	bne ~ne1
	ldx #0
	rts
~ne1
	ldx #1
	rts
;
; signed top stack item > xy
;
gt
	jsr ~cmp
	beq ~gt1
	bcc ~gt1
	ldx #1
	rts
~gt1
	ldx #0
	rts
;
; signed top stack item <= xy
;
le
	jsr ~cmp
	beq ~le1
	bcc ~le1
	ldx #0
	rts
~le1
	ldx #1
	rts
;
; signed top stack item >= xy
;
ge
	jsr ~cmp
	bcs ~ge1
	ldx #0
	rts
~ge1
	ldx #1
	rts
;
; signed top stack item < xy
;
lt
	jsr ~cmp
	bcc ~lt1
	ldx #0
	rts
~lt1
	ldx #1
	rts
;
; signed compare top stack item and xy
;
~cmp
	jsr pop
	stx pr
	tya
	ldy #0		;result hi
	eor #$80
	sta pr+1
	lda sr+1
	eor #$80
	cmp pr+1
	bne ~cmp1
	lda sr
	cmp pr
~cmp1
	rts
;
; unsigned top stack item > xy
;
ugt
	jsr ~ucmp
	beq ~ugt1
	bcc ~ugt1
	ldx #1
	rts
~ugt1
	ldx #0
	rts
;
; unsigned top stack item <= xy
;
ule
	jsr ~ucmp
	beq ~ule1
	bcc ~ule1
	ldx #0
	rts
~ule1
	ldx #1
	rts
;
; unsigned top stack item >= xy
;
uge
	jsr ~ucmp
	bcs ~uge1
	ldx #0
	rts
~uge1
	ldx #1
	rts
;
; unsigned top stack item < xy
;
ult
	jsr ~ucmp
	bcc ~ult1
	ldx #0
	rts
~ult1
	ldx #1
	rts
;
; unsigned compare top stack item and xy
;
~ucmp
	jsr pop
	stx pr
	sty pr+1
	ldy #0		;result hi
	lda sr+1
	cmp pr+1
	bne ~ucmp1
	lda sr
	cmp pr
~ucmp1
	rts
;
