/* rm.c    -  Remove files */

/* 07-Apr-1989  v0.70 - A.J.Travis */
/* 22-Jun-1991  v0.71 - J.G.Harston
 *   main() ends with exit(0);
 */

#include <stdio.h>

#ifdef MSDOS

#include <dos.h>
#define PLAIN 0x21		/* plain file [read-only] [archive] */
#define CFLAG 1		/* 8086 carry flag */

#else

#define W_CAT 1		/* write MOS catalogue info */
#define MODE 0			/* file mode */
#define ATTRIB 7		/* file attributes */
#define LOCKED 0x88		/* owner + other delete locks */
#define UNLOCK 0x77		/* permit owner + other delete */
#define EXISTS 0		/* file exists */
#define PLAIN 1		/* plain file */
#define DIRECTORY 2		/* directory file */
int fcb[9];			/* file control block */

#endif

#define ERROR -1
#define OK 0
#define SAME 0			/* strcmp() strings equal */

int force;			/* force removal of locked files */

int main(argc, argv)
int argc;
char *argv[];
{
	if (argc == 1)
		usage(-1);
	force = 0;
	while ((*argv[1] & 0xFF) == '-') {		/* BUG in compiler */
		if (strcmp(argv[1], "-f") == SAME)
			force++;
		else
			usage(-1);
		--argc;
		argv++;
	}
	while (--argc > 0) {
		if (force)
			kill(argv[1]);
		else
			delete(argv[1]);
		argv++;
	}
	exit(0);
}

/*
 * kill file
 * ---------
 * delete even if locked, and don't complain about non-existent files
 */
kill(file)
char *file;				/* name of file to delete */
{
#ifdef MSDOS
	union REGS reg;
	
	reg.h.ah = 0x43;		/* MSDOS chmod function */
	reg.h.al = 0x01;		/* set attribute mode */
	reg.x.cx = 0x00;		/* 'normal' attributes */
	reg.x.dx = file;		/* address of filename */

	/* Carry set indicates an error */
	if ((intdos(&reg, &reg) & CFLAG) == OK)
		unlink(file);
#else
	if (stat(file, fcb) == EXISTS) {
		if (fcb[MODE] & PLAIN) {		
			fcb[ATTRIB] = fcb[ATTRIB] & UNLOCK;
			osfile(file, fcb, W_CAT);	/* unlock file */
			unlink(file);
		}
	}
#endif
}

/*
 * delete file
 */
delete(file)
char *file;				/* name of file to delete */
{
#ifdef MSDOS
	if (findfirst(file, PLAIN) == NULL)
		fprintf(stderr, "rm: %s not found\n", file);
	else if (unlink(file) == ERROR)
		fprintf(stderr, "rm: %s permission denied\n", file);
#else
	if (stat(file, fcb) != EXISTS)
		fprintf(stderr, "rm: %s not found\n", file);
	else if (fcb[MODE] & PLAIN) {
		if (fcb[ATTRIB] & LOCKED)
			fprintf(stderr, "rm: %s permission denied\n", file);
		else
			unlink(file);
	}
	else if (fcb[MODE] & DIRECTORY)
		fprintf(stderr, "rm: can't remove directory %s\n", file);
	else
		fprintf(stderr, "rm: %s file type error\n", file);
#endif
}

/*
 * report correct usage, and exit
 */
usage(stat)
int stat;				/* exit status */
{
	fprintf(stderr, "usage: rm [-f] file1 ... [filen]\n");
	exit(stat);
}

