; sys.s  07-Apr-88  A.J.Travis

;
; Operating system interface for tcc under Acorn/BBC MOS
;
; 20-Jun-1991  v0.71 - J.G.Harston
;   system/exit pass return value via osbyte 1

;
; escape trap
;
escape
	cmp #6		;escape event?
	bne ~escape1
	lda #13		;disable escape event
	ldx #6		;event no.
	jsr osbyte
	lda #$7e	;acknowledge escape condition
	jsr osbyte
	lda #0		;close all open files
	ldy #0
	jsr osfind
	lda #$b2	;write keyboard semaphore
	ldy #$00	;and y
	ldx #$ff	;eor x (enable keyboard interrupts)
	jsr osbyte
	brk
	.byte $11	;escape 'error' code
	.byte $0a,$0d	;lf/cr
	.text "escape"
	.byte 0
~escape1
	rts
;
; exit()
; ------
; exit from user program
;
_exit
	lda #1		; return value is in xy
	jsr osbyte	; set return value
	ldx rsp		; get stored SP
	txs		; restore stack
	rts		; and return
;
; open(name, rwmode)
; ------------------
; exit with error status if file does not exist
; new files are opened with creat()
; MOS modes: read=$40, write=$80, r/w=$c0
;
_open
	jsr ~gsconv	;convert filename
	lda #<gsbuf	;MOS filename
	sta param	;file control block
	lda #>gsbuf
	sta param+1
	lda #5		;read catalogue information
	ldx #<param
	ldy #>param
	jsr osfile
	cmp #0
	beq ~openerr	;if zero, file does not exist
;
	ldy #2
	lda (sp),y	;rwmode lo
	clc
	adc #1		;convert to MOS format
	ror a
	ror a
	ror a
	ldx #<gsbuf	;filename
	ldy #>gsbuf
	jsr osfind
	tax		;return fd in xy
	beq ~openerr	;unless zero file handle
;
	ldy #0
	rts
~openerr
	ldx #$ff	;return -1 on error
	ldy #$ff
	rts
;
; creat(name, pmode)
; ------------------
; pmode ignored in this version ...
;
_creat
	jsr ~gsconv
	lda #$80	;open in write mode
	ldx #<gsbuf	;filename
	ldy #>gsbuf
	jsr osfind
	tax		;return fd in xy
	beq ~createrr	;unless zero file handle
;
	ldy #0
	rts
~createrr
	ldx #$ff	;return -1 on error
	ldy #$ff
	rts
;
; close(fd)
;
_close
	ldy #0
	lda (sp),y	;fd lo
	tay		;MOS fd parameter
	lda #0		;close file operation
	jsr osfind
	rts
;
; unlink(name)
;
_unlink
	jsr ~gsconv
	lda #<gsbuf	;filename
	sta param	;control block
	lda #>gsbuf
	sta param+1
	lda #6		;delete file
	ldx #<param
	ldy #>param
	jsr osfile
	rts
;
; stat(name, fcb)
;
_stat
	lda #5		;read catalogue information
	jmp ~osfile
;
; system(string)
; --------------
; execute MOS command
;
_system
	jsr ~gsconv	; convert to MOS format
	ldx #<gsbuf
	ldy #>gsbuf
	jsr oscli
	lda #1		; recover return value
	ldx #0		; clear it once read
	jsr osbyte	; read return value into x
	ldy #0		; clear y
	rts		; return return value in xy
;
; export address of MOS command line to C environment
;
__cmdlin
	lda #1		;get address of MOS command line		
	ldx #param	;4 byte parameter block
	ldy #0		;no file descriptor involved
	jsr osargs
	ldx param	;return address of '\r' terminated line
	ldy param+1
	rts
;
; convert string from C to MOS format
; -----------------------------------
; string is null terminated in c, $0d terminated in MOS
; use pr to copy filename from addr pointed to by
; top stack item into MOS general string input buffer
;
~gsconv
	ldy #0
	lda (sp),y	;name lo
	sta pr
	iny
	lda (sp),y	;name hi
	sta pr+1
	ldy #0
~conv
	lda (pr),y
	sta gsbuf,y	;parameter block for call
	beq ~term
	iny
	bne ~conv	;max length = 255 chars
~term
	lda #13		;replaces C '\0'
	sta gsbuf,y
	rts
;
; read(fd, buf, count)
; --------------------
; checks eof on fd before attempting to read
;
_read
	ldy #0
	lda (sp),y	;fd
	tax		;MOS fd in x
	lda #127	;check eof operation
	jsr osbyte
	txa		;exit status in x
	beq ~doread	;x=$ff is eof
	ldx #0		;return 0
	ldy #0
	rts
~doread
	lda #4
	pha
	jmp ~access
;
; write(fd, buf, count)
;
_write
	ldy #0
	lda (sp),y	;fd
	cmp #1		;fd 1 is stdout
	beq ~sput
	cmp #2		;fd 2 is stderr
	beq ~sput
	lda #2
	pha
	jmp ~access
;
; string output on stdout or stderr
;
~sput
	ldy #2
	lda (sp),y	;buf lo
	sta pr
	iny
	lda (sp),y	;buf hi
	sta pr+1
	iny
	lda (sp),y	;count lo
	sta tr
	iny
	lda (sp),y	;count hi
	tax
	ldy #0
~sput1
	lda tr
	bne ~sput2	 	
	cpx #0
	beq ~sput3
	dex
~sput2
	dec tr
	lda (pr),y
	jsr ~aput
	iny
	bne ~sput1
	inc pr+1
	jmp ~sput1
~sput3
	rts	
;
; MOS file access interface to tcc
; --------------------------------
; on entry:
;	a=2: write, ignoring new file ptr
;	a=4: read, ignoring new file ptr
;
; high order buffer address is zero, so the 6502 second processor
; accesses the 'correct' buffer in second processor memory.
;
~access
	ldy #0
	lda (sp),y	;fd lo
	sta param
	ldy #2
	lda (sp),y	;buf addr lo
	sta param+1
	iny
	lda (sp),y	;buf addr hi
	sta param+2
	lda #0		;hi order buf addr bytes
	sta param+3
	sta param+4
	iny
	lda (sp),y	;count lo
	sta param+5
	iny
	lda (sp),y	;count hi
	sta param+6
	lda #0
	sta param+7	;zero high order count
	sta param+8
	pla		;get access mode
	ldx #<param	;MOS control block
	ldy #>param
	jsr osgbpb
	sec		;calc no. transfers
	ldy #4
	lda (sp),y	;count lo
	sbc param+5	;resid lo
	tax		;ntransfers lo
	iny
	lda (sp),y	;count hi
	sbc param+6	;resid hi
	tay		;ntransfers hi
	rts
;
; getc(fd)
;
_getc
	ldy #0
	lda (sp),y	;fd lo
	beq ~getchar	;fd 0 is stdin
	tay		;MOS fd in y
	jsr osbget
	bcs ~geof	;c indicates eof on fd
	tax
	jmp ext		;sign extend char
~geof
	ldx #$ff	;return -1 on eof
	ldy #$ff
	rts
;
; input char from keyboard/rs423
;
~getchar
	jsr osrdch
	cmp #$0d	;\r
	bne ~echo
	jsr oswrch	;echo \r on input
	lda #$0a	;map \r to \n on input
~echo
	jsr oswrch	;echo chars on input
	tax
	jmp ext		;sign extend char
;
; putc(c, fd)
;
_putc
	ldy #0
	lda (sp),y	;char to put
	tax		;save in x for now
	ldy #2
	lda (sp),y	;fd
	cmp #1		;fd 1 is stdout
	beq ~xput
	cmp #2		;fd 2 is stderr
	beq ~xput
	tay		;MOS fd in y
	txa		;MOS char in a
	jsr osbput
	rts
;
; output x to screen/rs423
;
~xput
	txa
;
; output a to screen/rs423
;
~aput
	cmp #$0a	;\n
	beq ~mapn
	jsr oswrch
	rts
;
; map \n to \r\n on output
;
~mapn
	lda #$0d	;\r
	jsr oswrch
	lda #$0a	;\n
	jsr oswrch
	rts
;
; vdu(c)
; ------
; output char to keyboard/rs423 without \n to \r\n transformation
;
_vdu
	ldy #0
	lda (sp),y
	jsr oswrch
	rts
;
; osbyte(type, parameters)
; ------------------------
; MOS "osbyte" interface
; a is type of call
; x and y are parameters
;
_osbyte
	ldy #2
	lda (sp),y	;x parameter
	tax
	iny
	lda (sp),y	;y parameter
	sta asave	;save it
	ldy #0
	lda (sp),y	;type of call
	ldy asave
	jsr osbyte
	rts		;return val in xy
;
; osword(type, address)
; ---------------------
; MOS "osword" interface
; a is type of call
; (x + y << 8) is address of parameter block
;
_osword
	ldy #2
	lda (sp),y	;address lo
	tax
	iny
	lda (sp),y	;address hi
	sta asave	;save it
	ldy #0
	lda (sp),y	;type of call
	ldy asave
	jsr osword
	tya
	tax		;return y in low byte
	php
	pla
	and #$01
	tay		;return carry in high byte
	rts		;return val in xy
;
; osfile(name, fcb, type)
; -----------------------
; MOS osfile interface
;
_osfile
	ldy #4
	lda (sp),y	;osfile type
~osfile	
	pha		;save type
	jsr ~gsconv	;convert name to MOS format
	ldy #2
	lda (sp),y	;fcb address lo
	sta tr
	iny
	lda (sp),y	;fcb address hi
	sta tr+1
	ldy #0
	lda #<gsbuf	;update file control block
	sta (tr),y	;fcb address lo
	iny
	lda #>gsbuf
	sta (tr),y	;fcb address hi
	pla		;osfile type
	ldx tr		;addr of fcb
	ldy tr+1
	jsr osfile
	cmp #0		;check MOS exit status, 0 = no file
	bne ~osfile1
	ldx #$ff	;return -1 if no file found
	ldy #$ff
	rts
~osfile1
	ldx #0		;return 0 otherwise
	ldy #0
	sta (tr),y	;leave MOS file type in fcb
	rts
;
; isalpha(c)
;
_isalpha
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$03	;_U | _L
	tax
	beq ~isalpha1
	ldx #$01	;return true
~isalpha1
	rts		;result in x
;
; isupper(c)
;
_isupper
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$01	;_U
	tax
	beq ~isupper1
	ldx #$01	;return true
~isupper1
	rts		;result in x
;
; islower(c)
;
_islower
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$02	;_L
	tax
	beq ~islower1
	ldx #$01	;return true
~islower1
	rts		;result in x
;
; isdigit(c)
;
_isdigit
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$04	;_N
	tax
	beq ~isdigit1
	ldx #$01	;return true
~isdigit1
	rts		;result in x
;
; isxdigit(c)
;
_isxdigi
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$44	;_N | _X
	tax
	beq ~isxdigit1
	ldx #$01	;return true
~isxdigit1
	rts		;result in x
;
; isspace(c)
;
_isspace
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$08	;_S
	tax
	beq ~isspace1
	ldx #$01	;return true
~isspace1
	rts		;result in x
;
; ispunct(c)
;
_ispunct
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$10	;_P
	tax
	beq ~ispunct1
	ldx #$01	;return true
~ispunct1
	rts		;result in x
;
; isalnum(c)
;
_isalnum
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$07	;_U | _L | _N
	tax
	beq ~isalnum1
	ldx #$01	;return true
~isalnum1
	rts		;result in x
;
; isprint(c)
;
_isprint
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$17	;_P | _U | _L | _N
	tax
	beq ~isprint1
	ldx #$01	;return true
~isprint1
	rts		;result in x
;
; iscntrl(c)
;
_iscntrl
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$20	;_C
	tax
	beq ~iscntrl1
	ldx #$01	;return true
~iscntrl1
	rts		;result in x
;
; isascii(c)
;
_isascii
	ldy #0
	lda (sp),y	;low byte of c
	and #$80	;0 if <= 127
	eor #$80	;invert
	tax
	beq ~isascii1
	ldx #$01	;return true
~isascii1
	rts		;result in x
;
; toupper(c)
;
_toupper
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$02	;_L
	beq ~toupper1	;skip if not lower-case
	sec
	txa		;original char
	sbc #'a' - 'A'	;force upper case
	tax
~toupper1
	rts		;result in x
;
; tolower(c)
;
_tolower
	ldy #0
	lda (sp),y	;low byte of c
	tax
	lda _ctype_,x
	and #$01	;_U
	beq ~tolower1	;skip if not upper-case
	clc
	txa		;original char
	adc #'a' - 'A'	;force lower case
	tax
~tolower1
	rts		;result in x
;
; toascii(c)
;
_toascii
	ldy #0
	lda (sp),y	;low byte of c
	and #$7f
	tax
	rts		;result in x
;
; character classification table
;
	.byte $00	;EOF 'subscript' is -1
_ctype_
	.byte $20,$20,$20,$20,$20,$20,$20,$20
	.byte $20,$08,$08,$08,$08,$08,$20,$20
	.byte $20,$20,$20,$20,$20,$20,$20,$20
	.byte $20,$20,$20,$20,$20,$20,$20,$20
	.byte $18,$10,$10,$10,$10,$10,$10,$10
	.byte $10,$10,$10,$10,$10,$10,$10,$10
	.byte $04,$04,$04,$04,$04,$04,$04,$04
	.byte $04,$04,$10,$10,$10,$10,$10,$10
	.byte $10,$41,$41,$41,$41,$41,$41,$01
	.byte $01,$01,$01,$01,$01,$01,$01,$01
	.byte $01,$01,$01,$01,$01,$01,$01,$01
	.byte $01,$01,$01,$10,$10,$10,$10,$10
	.byte $10,$42,$42,$42,$42,$42,$42,$02
	.byte $02,$02,$02,$02,$02,$02,$02,$02
	.byte $02,$02,$02,$02,$02,$02,$02,$02
	.byte $02,$02,$02,$10,$10,$10,$10,$20
;
